-- by modelleicher
-- www.schwabenmodding.de
-- facebook.de/modelleicher
-- by modelleicher
-- December 2016
-- www.schwabenmodding.de
-- facebook.de/modelleicher

-- Version 1.0, Initial version 
-- Version 1.1, MP Fix 29-01-2017 - all tree attachers in the same table within "self" 




treeAttacherNodes = {};

function treeAttacherNodes.prerequisitesPresent(specializations)
    return true;
end;

function treeAttacherNodes:load(savegame)

	-- MP fix 29.01.2017 
	-- in case the needed tables aren't setup yet, setup:
	if self.winchUtilHelper == nil then -- global winch util helper
		self.winchUtilHelper = {}; 
		self.winchUtilHelper.treeAttacher = {}; -- all tree attachers will be in this table and used from this table
		self.winchUtilHelper.treeAttacherCount = 0; -- count of all tree attachers used within this mod
	end;
	
	self.localWUH_tAN = {}; -- local winch util helper
	self.localWUH_tAN.treeAttacherCountMin = self.winchUtilHelper.treeAttacherCount + 1; -- the min count is the current count + 1 

	
	self.tAN = {};
	
	
	self.tAN.attacherNodes = {};
	
	--self.tAN.playerRange = 0;
   
	local i = 0;
	while true do
		local an = {};
		local xmlKey = "vehicle.treeAttacherNodes.attacherNode("..i..")";
		local compCnt = getXMLInt(self.xmlFile, xmlKey.."#attacherComponent");
		if compCnt == nil then				
			break;
		end;
		an.attacherNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, xmlKey.."#attacherNode"));
		an.attacherCompIndex = compCnt;
		an.rope = Utils.indexToObject(self.components, getXMLString(self.xmlFile, xmlKey.."#rope"));
		an.numAttachable = Utils.getNoNil(getXMLInt(self.xmlFile, xmlKey.."#numAttachable"), 1);
		--an.rangeAttachable = Utils.getNoNil(getXMLInt(self.xmlFile, xmlKey.."#rangeAttachable"), 1);
		--self.tAN.playerRange = math.max(self.tAN.playerRange, an.rangeAttachable);
		
		setVisibility(an.rope, false);
		
		an.massTimer = false;
		an.massTimerMs = 0;
		
		-- MP Fix 29.01.2017
				
		-- first the attacher count +1, this is the index of the current attacher then
		self.winchUtilHelper.treeAttacherCount = self.winchUtilHelper.treeAttacherCount +1;
		
		-- save the given index within the table for further access just in case
		an.treeAttacherCountIndex = self.winchUtilHelper.treeAttacherCount;
		
		-- table of this attacher is inserted in winchUtilHelper attacher table at the given index
		self.winchUtilHelper.treeAttacher[self.winchUtilHelper.treeAttacherCount] = an;

		
		table.insert(self.tAN.attacherNodes, an);
		
		i = i+1;
		self.tANCnt = i;
	end;
   
	self.tAN.maxRange = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.treeAttacherNodes#maxRange"), 1.5);
	self.tAN.maxRangeRef = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.treeAttacherNodes#maxRangeRef"));
	
	
	self.tANCnt = table.maxn(self.tAN.attacherNodes);
	
	-- now we have all attachers in this spezialiazion loaded.. so the local max count is updated
	self.localWUH_tAN.treeAttacherCountMax = self.winchUtilHelper.treeAttacherCount;
	
	--self.tAN.playerRange = self.tAN.playerRange +1;
 
end;


-- basic concept
-- . use same features as winch
-- . attach to closest attacher node 
-- . only attach within attacher node range
-- . only attach if attacher node isn't "full" yet
-- 


function treeAttacherNodes:delete()
end;
function treeAttacherNodes:mouseEvent(posX, posY, isDown, isUp, button)
end;
function treeAttacherNodes:keyEvent(unicode, sym, modifier, isDown)
end

function treeAttacherNodes:update(dt)    
	local updated = false;
	--if self:getIsActive() then
		if self.isServer then
			for i=self.localWUH_tAN.treeAttacherCountMin, self.localWUH_tAN.treeAttacherCountMax do
				local an = self.winchUtilHelper.treeAttacher[i] -- MP Fix 30.01.2017 -- get the current attacher from attacher table
				if an.localAttachedSplit ~= nil then
					if not entityExists(an.localAttachedSplit) then -- split was deleted some way 
						an.serverAttachedSplit = nil;    
						an.localAttachedSplit = nil;
						an.localSplitAttacherRef = nil;
						an.foundLocalSplit = nil;
						an.wUpdated = true;
						an.winching = false;
						setVisibility(an.rope, false); 
						print("Split doesn't exist anymore");
					end;
						-- rope animation
					local x, y, z = getWorldTranslation(an.localSplitAttacherRef);
					local a, b, c = getWorldTranslation(an.rope);
					local dirX, dirY, dirZ = worldDirectionToLocal(getParent(an.attacherNode), x-a, y-b, z-c);
					local upx, upy, upz = 0,1,0;
					setDirection(an.rope, dirX, dirY, dirZ, upx, upy, upz);
					local distance = Utils.vector3Length(a-x, b-y, c-z);
					setScale(an.rope, 1, 1, distance);
					setVisibility(an.rope, true);
					
					
					-- FS17 Fix, add additional force to keep trees on the attacher since the Joints are.. meh.
					local aX, aY, aZ = getWorldTranslation(an.attacherNode);
					local actualDistance = Utils.vector3Length(x-aX, y-aY, z-aZ);
					if actualDistance > an.curLimit then -- if actual distance is longer than joint limit the joint is failing..
						-- the difference between actual distance and supposed to be distance are used to calculate the force needed
						local difference = actualDistance - an.curLimit;
						
						-- 
						-- 4 => base value; * difference^2 => linear increase isn't enough; *splitt mass => heavier trees need more force; 
						-- * curLimit => the larger the trans limit the less stable the joint gets, again linear increase isn't enough.
						local force = (4 * difference^2)*an.splitBackupMass*(an.curLimit^2);
						--renderText(0.1, 0.1, 0.02, "raw: "..tostring(force));
						-- cap the force to a certain max value to not have everything fly across the map in case a tree gets stuck
						force = math.max(math.min(force, 4*an.splitBackupMass*(an.curLimit^2)), -4*an.splitBackupMass*(an.curLimit^2));
						
		
						-- direction stuff:
						-- attacher split ref is x, y, z
						-- attaher node ref is aX, aY, aZ
						-- local direction relative to attacher split ref is:
						local dirX2 = aX - x;
						local dirY2 = aY - y;
						local dirZ2 = aZ - z;
						
						-- debug
						if Vehicle.debugRendering and self:getIsActiveForInput() then
												
							local ddx, ddy, ddz = localDirectionToLocal(an.localSplitAttacherRef, an.localAttachedSplit, dirX2/2, dirY2/2, dirZ2/2);
						
							local wx, wy, wz = localToWorld(an.localAttachedSplit, ddx, ddy, ddz)
						
							renderText(0.1, 0.12+(i*0.01), 0.02, "force "..tostring(i)..": - "..tostring(force));
							drawDebugPoint(wx, wy, wz, 1, 0, 0, 1);	
							drawDebugArrow(x, y, z, dirX2, dirY2, dirZ2, 0, 0, 1, 1, 1, 0);
						end;					

						
						-- add the force to the tree
						addForce(an.localAttachedSplit, dirX2*force,dirY2*force,dirZ2*force, x,y,z, false);
						
					end;					
				end;
				
				
			end;
		end;
	--end;
	
  
	  -- individual player
	if g_currentMission.controlPlayer and g_currentMission.player ~= nil then
			-- only on first update frame:
		if self.firstTimeActive == nil then -- add the search plane to player
			self.tAN.searchNodes = {};
			self.tAN.searchNodes.center = createTransformGroup("centerNode");
			link(g_currentMission.player.cameraNode, self.tAN.searchNodes.center);
			setTranslation(self.tAN.searchNodes.center, 0, 0, -1);
			local planePoint, dir1, dir2 = winchUtil:createFinderShape(self.tAN.searchNodes.center);
			self.tAN.searchNodes.planePoint = planePoint;
			self.tAN.searchNodes.dir1 = dir1;
			self.tAN.searchNodes.dir2 = dir2;
			self.firstTimeActive = true;
		end;
	
		-- get the distance from the player to the winch to check if the player is within the length of the rope
		local x,y,z = getWorldTranslation(self.tAN.searchNodes.center);
		local a,b,c = getWorldTranslation(self.tAN.maxRangeRef)
		local distance = Utils.vector3Length(x-a, y-b, z-c);
		if distance < self.tAN.maxRange then -- player is in attaching range
			
			-- find trees...
			local pX, pY, pZ = getWorldTranslation(self.tAN.searchNodes.planePoint);
			local dir1X, dir1Y, dir1Z = getWorldTranslation(self.tAN.searchNodes.dir1);
			local dir2X, dir2Y, dir2Z = getWorldTranslation(self.tAN.searchNodes.dir2);

			local nx, ny, nz, yx, yy, yz = dir1X-pX, dir1Y-pY, dir1Z-pZ, dir2X-pX, dir2Y-pY, dir2Z-pZ;
			
			local shape, minY, maxY, minZ, maxZ = findSplitShape(pX,pY,pZ, nx, ny, nz, yx, yy, yz, 1, 1);
			
			if shape ~= nil and shape ~= 0 then -- found split shape
				local currentLocalSplit = shape;
				local currentLocalSplitSizes = {minY, maxY, minZ, maxZ};
								
				local alreadyAttached = false;
				

				
				for i = self.localWUH_tAN.treeAttacherCountMin, self.localWUH_tAN.treeAttacherCountMax do
					local an = self.winchUtilHelper.treeAttacher[i];
					if an.localAttachedSplit == shape then
					
						alreadyAttached = true;
						
						if g_languageSuffix == "_de" then
							renderText(0.3, 0.0216, 0.02, "Baum angehängt. Taste X: Abhängen");
						else
							renderText(0.3, 0.0216, 0.02, "Tree attached. Press X to detach");
						end;
						
						-- check if detaching is enabled, detachTree
						if updated == false and InputBinding.hasEvent(InputBinding.TAN_ATTACH) then
							updated = true;
							winchUtil:clientServerDetachTree(self, an.treeAttacherCountIndex);
							--print("WINCH_DETACH_INPUT_CALL");
						end;    
					end;				
				end;
				if not alreadyAttached and updated == false then
					--print("not already attached");
					if g_languageSuffix == "_de" then
						renderText(0.3, 0.0216, 0.02, "Baum gefunden. Taste X zum anhängen");
					else
						renderText(0.3, 0.0216, 0.02, "Tree in Range. Press X to attach");
					end;
				
					-- attach the closest node
					if InputBinding.hasEvent(InputBinding.TAN_ATTACH) then -- attach log
						updated = true;
						--print("INPUT CALL");
						-- get the closest attacher node
						local closestNode = nil;
						local closestDistance = 1000;
						--print(tostring(self.tANCnt));
						for i = self.localWUH_tAN.treeAttacherCountMin, self.localWUH_tAN.treeAttacherCountMax do
							local an = self.winchUtilHelper.treeAttacher[i]
							--print("for loop"..tostring(i));
							if an.localAttachedSplit == nil then -- only attach to unused nodes
								local x1,y1,z1 = getWorldTranslation(self.tAN.searchNodes.center);
								local x2,y2,z2 = getWorldTranslation(an.attacherNode)
								local anDistance = Utils.vector3Length(x1-x2, y1-y2, z1-z2);
								
								if anDistance < closestDistance then -- if current node is closer than previous nodes, new closest node is current node
									closestDistance = anDistance;
									closestNode = an;
									--print("closer one found: "..tostring(closestDistance));
								end;
							end;
						end;
						
					
						if closestNode ~= nil then -- node does exist
							--print("attaching...");
							closestNode.currentLocalSplit = shape;
							closestNode.currentLocalSplitSizes = currentLocalSplitSizes;
						
							local tX, tY, tZ = getWorldTranslation(self.tAN.searchNodes.center);
							local rX, rY, rZ = getWorldRotation(self.tAN.searchNodes.center);
							
							winchUtil:clientServerFindTree(self, closestNode.treeAttacherCountIndex, tX, tY, tZ, rX, rY, rZ, true, nil, 0, 0, 0, 0, 0);
						end;
					end;
				end;
			else    
				for i=self.localWUH_tAN.treeAttacherCountMin, self.localWUH_tAN.treeAttacherCountMax do
					self.winchUtilHelper.treeAttacher[i].currentLocalSplit = nil;
				end;
			end;
			-- check if winching is enabled
			
		end;
	end;
end;    


function treeAttacherNodes:draw()
   
end;
                
